# Copyright 2023-2024 Vincent Jacques

from typing import Iterable
import unittest

import matplotlib.pyplot

from .classification import Problem, Model, Alternative


def visualize_classification_model(problem: Problem, model: Model, alternatives: Iterable[Alternative], axes: matplotlib.pyplot.Axes):
    """
    Create a visual representation of a classification model and classified alternatives, using Matplotlib.
    """

    # @todo(Feature, v1.2) Visualize single-peaked criteria.
    # See the possible solutions in e-mail "Visualisation des critères single-peaked" 2023-11-24

    criteria_count = len(problem.criteria)
    assert criteria_count >= 1
    categories_count = len(problem.ordered_categories)
    assert categories_count >= 2
    boundaries_count = categories_count - 1

    vertical_margin = 0.05
    horizontal_margin = 0.1

    if criteria_count == 1:
        xs = [0.5]
    else:
        xs = [horizontal_margin + criterion_index * (1 - 2 * horizontal_margin) / (criteria_count - 1) for criterion_index in range(criteria_count)]

    if criteria_count <= 2:
        criterion_index_for_alternatives = 0
    else:
        criterion_index_for_alternatives = 1

    axes.set_xlim(0, 1)
    axes.set_xticks(xs, [criterion.name for criterion in problem.criteria])

    axes.set_ylim(-vertical_margin, 1 + vertical_margin)
    axes.set_yticks([0, 1], ["worst", "best"])

    boundary_profiles = [[] for _ in problem.ordered_categories[1:]]
    for criterion, accepted_values in zip(problem.criteria, model.accepted_values):
        assert accepted_values.is_thresholds
        if criterion.is_real:
            for boundary_index in range(boundaries_count):
                boundary_profiles[boundary_index].append(accepted_values.real_thresholds.thresholds[boundary_index])
        elif criterion.is_integer:
            for boundary_index in range(boundaries_count):
                boundary_profiles[boundary_index].append(accepted_values.integer_thresholds.thresholds[boundary_index])
        else:
            assert criterion.is_enumerated
            for boundary_index in range(boundaries_count):
                boundary_profiles[boundary_index].append(criterion.enumerated_values.get_value_rank(accepted_values.enumerated_thresholds.thresholds[boundary_index]))

    def extend(ys):
        return [ys[0]] + ys + [ys[-1]]
    ys = [
        extend(normalize_profile(problem.criteria, boundary_profile))
        for boundary_profile in boundary_profiles
    ]
    ys.append(extend([1] * len(xs)))
    unstacked_ys = [ys[0]]
    for ys1, ys2 in zip(ys[1:], ys[:-1]):
        unstacked_ys.append([y1 - y2 for y1, y2 in zip(ys1, ys2)])
    collections = axes.stackplot([0] + xs + [1], unstacked_ys, alpha=0.4)
    colors = [collection.get_facecolor()[0] for collection in collections]

    for (x, criterion) in zip(xs, problem.criteria):
        criterion_axis = axes.secondary_yaxis(x)
        if criterion.is_real:
            values = criterion.real_values
            ticks = [0, 0.25, 0.5, 0.75, 1]
            labels = [f"{values.min_value + tick * (values.max_value - values.min_value):.1f}" for tick in ticks]
            if values.is_increasing:
                criterion_axis.set_yticks(ticks, labels)
            else:
                assert values.is_decreasing
                criterion_axis.set_yticks(ticks, reversed(labels))
        elif criterion.is_integer:
            values = criterion.integer_values
            labels = list(make_integer_labels(values.min_value, values.max_value))
            ticks = [(label - values.min_value) / (values.max_value - values.min_value) for label in labels]
            if values.is_increasing:
                criterion_axis.set_yticks(ticks, labels)
            else:
                assert values.is_decreasing
                criterion_axis.set_yticks(ticks, reversed(labels))
        else:
            assert criterion.is_enumerated
            values = criterion.enumerated_values
            ticks_count = len(values.ordered_values)
            ticks = [n / (ticks_count - 1) for n in range(ticks_count)]
            criterion_axis.set_yticks(ticks, values.ordered_values)

    for (category_index, category) in enumerate(problem.ordered_categories):
        if category_index == 0:
            low_y = 0
        else:
            low_y = normalize_value(problem.criteria[0], boundary_profiles[category_index - 1][0])
        if category_index == len(problem.ordered_categories) - 1:
            high_y = 1
        else:
            high_y = normalize_value(problem.criteria[0], boundary_profiles[category_index][0])
        y = (low_y + high_y) / 2
        color = colors[category_index]
        axes.text(
            0, y,
            category.name,
            color=color,
            alpha=1,
            fontweight="bold", verticalalignment="center"
        )

    for alternative in alternatives:
        if alternative.category_index is None:
            color = "black"
        else:
            color = colors[alternative.category_index]
        axes.plot(
            xs, normalize_profile(problem.criteria, make_numeric_profile(problem.criteria, alternative.profile)),
            "o--",
            label=alternative.name,
            color=color,
            alpha=1,
        )
        axes.text(
            xs[criterion_index_for_alternatives],
            normalize_value(
                problem.criteria[criterion_index_for_alternatives],
                make_numeric_value(
                    problem.criteria[criterion_index_for_alternatives],
                    alternative.profile[criterion_index_for_alternatives],
                ),
            ),
            alternative.name,
            color=color, alpha=1)


def make_integer_labels(min_value, max_value):
    intervals_count = max_value - min_value
    assert intervals_count >= 1
    if intervals_count == 1:
        yield min_value
        yield max_value
    elif intervals_count == 2:
        yield min_value
        yield min_value + 1
        yield max_value
    elif intervals_count % 2 == 0:
        yield min_value
        yield min_value + intervals_count // 4
        yield (min_value + max_value) // 2
        yield max_value - intervals_count // 4
        yield max_value
    else:
        yield min_value
        yield min_value + intervals_count // 3
        yield max_value - intervals_count // 3
        yield max_value


class MakeIntegerLabelsTests(unittest.TestCase):
    def _test(self, min_value, max_value, expected):
        self.assertEqual(list(make_integer_labels(min_value, max_value)), expected)

    def test_all(self):
        for n in range(100):
            l = list(make_integer_labels(10, n + 11))
            self.assertLessEqual(len(l), 5)
            self.assertEqual(l[0], 10)
            self.assertEqual(l[-1], n + 11)
            if len(l) % 2 == 1:
                self.assertEqual(l[len(l) // 2], (10 + n + 11) // 2)
            self.assertEqual(l, sorted(l))

    def test_0_1(self):
        self._test(0, 1, [0, 1])

    def test_0_2(self):
        self._test(0, 2, [0, 1, 2])

    def test_0_3(self):
        self._test(0, 3, [0, 1, 2, 3])

    def test_0_4(self):
        self._test(0, 4, [0, 1, 2, 3, 4])

    def test_10_14(self):
        self._test(10, 14, [10, 11, 12, 13, 14])

    def test_0_5(self):
        self._test(0, 5, [0, 1, 4, 5])

    def test_0_6(self):
        self._test(0, 6, [0, 1, 3, 5, 6])

    def test_0_7(self):
        self._test(0, 7, [0, 2, 5, 7])

    def test_0_70(self):
        self._test(0, 70, [0, 17, 35, 53, 70])

    def test_0_8(self):
        self._test(0, 8, [0, 2, 4, 6, 8])

    def test_0_8(self):
        self._test(0, 80, [0, 20, 40, 60, 80])

    def test_bug_1(self):
        self._test(-4949, 9942, [-4949, 14, 4979, 9942])

    def test_0_9000(self):
        self._test(0, 9000, [0, 2250, 4500, 6750, 9000])

    def test_0_9001(self):
        self._test(0, 9001, [0, 3000, 6001, 9001])

    def test_0_9002(self):
        self._test(0, 9002, [0, 2250, 4501, 6752, 9002])

    def test_0_9003(self):
        self._test(0, 9003, [0, 3001, 6002, 9003])

    def test_0_9004(self):
        self._test(0, 9004, [0, 2251, 4502, 6753, 9004])


def make_numeric_profile(criteria, profile):
    return [
        make_numeric_value(criterion, y)
        for (criterion, y) in zip(criteria, profile)
    ]


def make_numeric_value(criterion, y):
    if criterion.is_real:
        return y.real.value
    elif criterion.is_integer:
        return y.integer.value
    else:
        assert criterion.is_enumerated
        return criterion.enumerated_values.get_value_rank(y.enumerated.value)


def normalize_profile(criteria, ys):
    return [
        normalize_value(criterion, y)
        for (criterion, y) in zip(criteria, ys)
    ]


def normalize_value(criterion, y):
    is_increasing = None
    if criterion.is_real:
        values = criterion.real_values
        is_increasing = values.is_increasing
        y = (y - values.min_value) / (values.max_value - values.min_value)
    elif criterion.is_integer:
        values = criterion.integer_values
        is_increasing = values.is_increasing
        y = (y - values.min_value) / (values.max_value - values.min_value)
    else:
        assert criterion.is_enumerated
        values = criterion.enumerated_values
        is_increasing = True
        y = y / (len(values.ordered_values) - 1)

    if is_increasing:
        return y
    else:
        return 1 - y



from typing import Iterable
import unittest

from .classification import Problem, Criterion, Category, Model, AcceptedValues, SufficientCoalitions


def describe_classification_problem(problem: Problem) -> Iterable[str]:
    """
    Generate a human-readable description of a classification problem.
    """

    categories_count = len(problem.ordered_categories)
    assert categories_count >= 2
    criteria_count = len(problem.criteria)
    assert criteria_count >= 1

    category_names = [f'"{category.name}"' for category in problem.ordered_categories]
    category_names_joined = ", ".join(category_names[:-1]) + " and " + category_names[-1]
    yield f"This a classification problem into {categories_count} ordered categories named {category_names_joined}."
    yield f"The best category is {category_names[-1]} and the worst category is {category_names[0]}."

    if criteria_count == 1:
        yield "There is 1 classification criterion."
    else:
        yield f"There are {criteria_count} classification criteria (in no particular order)."
    for criterion in problem.criteria:
        if criterion.is_real:
            values = criterion.real_values
            yield f'Criterion "{criterion.name}" takes real values between {values.min_value:.1f} and {values.max_value:.1f} included.'
            if values.is_increasing:
                yield f'Higher values of "{criterion.name}" are known to be better.'
            else:
                yield f'Lower values of "{criterion.name}" are known to be better.'
        elif criterion.is_integer:
            values = criterion.integer_values
            yield f'Criterion "{criterion.name}" takes integer values between {values.min_value} and {values.max_value} included.'
            if values.is_increasing:
                yield f'Higher values of "{criterion.name}" are known to be better.'
            else:
                yield f'Lower values of "{criterion.name}" are known to be better.'
        else:
            assert criterion.is_enumerated
            values = criterion.enumerated_values
            yield f'Criterion "{criterion.name}" takes values in the following set: {", ".join(f"{value}" for value in values.ordered_values)}.'
            yield f'The best value for criterion "{criterion.name}" is "{values.ordered_values[-1]}" and the worst value is "{values.ordered_values[0]}".'


class DescribeClassificationProblemTestCase(unittest.TestCase):
    maxDiff = None

    def _test(self, problem, expected):
        self.assertEqual(list(describe_classification_problem(problem)), expected)

    def test_simplest(self):
        self._test(
            Problem(
                [
                    Criterion("Criterion", Criterion.RealValues(Criterion.PreferenceDirection.increasing, 0, 1)),
                ],
                [Category("Bad"), Category("Good")],
            ),
            [
                'This a classification problem into 2 ordered categories named "Bad" and "Good".',
                'The best category is "Good" and the worst category is "Bad".',
                'There is 1 classification criterion.',
                'Criterion "Criterion" takes real values between 0.0 and 1.0 included.',
                'Higher values of "Criterion" are known to be better.'
            ]
        )

    def test_many_categories(self):
        self._test(
            Problem(
                [
                    Criterion("Criterion", Criterion.RealValues(Criterion.PreferenceDirection.increasing, 0, 1)),
                ],
                [Category("Worsestest"), Category("Interm 1"), Category("Interm 2"), Category("Interm 3"), Category("Bestestest")],
            ),
            [
                'This a classification problem into 5 ordered categories named "Worsestest", "Interm 1", "Interm 2", "Interm 3" and "Bestestest".',
                'The best category is "Bestestest" and the worst category is "Worsestest".',
                'There is 1 classification criterion.',
                'Criterion "Criterion" takes real values between 0.0 and 1.0 included.',
                'Higher values of "Criterion" are known to be better.',
            ]
        )

    def test_criteria_variety(self):
        self._test(
            Problem(
                [
                    Criterion("Increasing real criterion", Criterion.RealValues(Criterion.PreferenceDirection.increasing, -5.2, 10.3)),
                    Criterion("Decreasing real criterion", Criterion.RealValues(Criterion.PreferenceDirection.decreasing, 5, 15)),
                    Criterion("Increasing integer criterion", Criterion.IntegerValues(Criterion.PreferenceDirection.increasing, 0, 10)),
                    Criterion("Decreasing integer criterion", Criterion.IntegerValues(Criterion.PreferenceDirection.decreasing, 4, 16)),
                    Criterion("Enumerated criterion", Criterion.EnumeratedValues(["A", "B", "C"])),
                ],
                [Category("Bad"), Category("Good")],
            ),
            [
                'This a classification problem into 2 ordered categories named "Bad" and "Good".',
                'The best category is "Good" and the worst category is "Bad".',
                'There are 5 classification criteria (in no particular order).',
                'Criterion "Increasing real criterion" takes real values between -5.2 and 10.3 included.',
                'Higher values of "Increasing real criterion" are known to be better.',
                'Criterion "Decreasing real criterion" takes real values between 5.0 and 15.0 included.',
                'Lower values of "Decreasing real criterion" are known to be better.',
                'Criterion "Increasing integer criterion" takes integer values between 0 and 10 included.',
                'Higher values of "Increasing integer criterion" are known to be better.',
                'Criterion "Decreasing integer criterion" takes integer values between 4 and 16 included.',
                'Lower values of "Decreasing integer criterion" are known to be better.',
                'Criterion "Enumerated criterion" takes values in the following set: A, B, C.',
                'The best value for criterion "Enumerated criterion" is "C" and the worst value is "A".',
            ]
        )


def describe_classification_model(problem: Problem, model: Model) -> Iterable[str]:
    """
    Generate a human-readable description of a classification model.
    """

    criteria_count = len(problem.criteria)
    assert len(model.accepted_values) == criteria_count
    assert criteria_count > 0
    categories_count = len(problem.ordered_categories)
    boundaries_count = categories_count - 1
    assert boundaries_count > 0
    assert len(model.sufficient_coalitions) == boundaries_count

    def comma_and(s):
        s = list(s)
        if len(s) == 1:
            return s[0]
        elif len(s) == 2:
            return " and ".join(s)
        else:
            return ", ".join(s[:-1]) + ", and " + s[-1]  # https://en.wikipedia.org/wiki/Serial_comma

    def make_upset_roots(upset_roots):
        for coalition in upset_roots:
            criterion_names = []
            for criterion_index in coalition:
                criterion = problem.criteria[criterion_index]
                criterion_names.append(f'"{criterion.name}"')
            yield f'  - {comma_and(criterion_names)}'

    def make_profile(accepted_values, boundary_index):
        for criterion_index, criterion in enumerate(problem.criteria):
            assert accepted_values[criterion_index].is_thresholds
            if criterion.is_real:
                assert len(accepted_values[criterion_index].real_thresholds.thresholds) == boundaries_count
                values = criterion.real_values
                constraint = "at least" if values.is_increasing else "at most"
                yield f'{constraint} {accepted_values[criterion_index].real_thresholds.thresholds[boundary_index]:.2f} on criterion "{criterion.name}"'
            elif criterion.is_integer:
                assert len(accepted_values[criterion_index].integer_thresholds.thresholds) == boundaries_count
                values = criterion.integer_values
                constraint = "at least" if values.is_increasing else "at most"
                yield f'{constraint} {accepted_values[criterion_index].integer_thresholds.thresholds[boundary_index]} on criterion "{criterion.name}"'
            else:
                assert criterion.is_enumerated
                assert len(accepted_values[criterion_index].enumerated_thresholds.thresholds) == boundaries_count
                yield f'at least "{accepted_values[criterion_index].enumerated_thresholds.thresholds[boundary_index]}" on criterion "{criterion.name}"'

    is_uc = all(sufficient_coalitions == model.sufficient_coalitions[0] for sufficient_coalitions in model.sufficient_coalitions[1:])
    if is_uc:
        first_sufficient_coalitions = model.sufficient_coalitions[0]
        if first_sufficient_coalitions.is_weights:
            yield "This is a MR-Sort (a.k.a. 1-Uc-NCS) model: an NCS model where the sufficient coalitions are specified using the same criterion weights for all boundaries."
            yield "The weights associated to each criterion are:"
            assert len(first_sufficient_coalitions.weights.criterion_weights) == criteria_count
            for criterion, weight in zip(problem.criteria, first_sufficient_coalitions.weights.criterion_weights):
                yield f'  - Criterion "{criterion.name}": {weight:.2f}'
            yield "To get into an upper category, an alternative must be better than the following profiles on a set of criteria whose weights add up to at least 1:"
        else:
            assert first_sufficient_coalitions.is_roots
            yield "This is a Uc-NCS model: an NCS model with the same sufficient coalitions for all boundaries."
            yield "The sufficient coalitions of criteria are the following, as well as any of their unions:"
            yield from make_upset_roots(first_sufficient_coalitions.roots.upset_roots)
            yield "To get into an upper category, an alternative must be better than the following profiles on a sufficient coalition of criteria:"
        for boundary_index, category in enumerate(problem.ordered_categories[1:]):
            yield f'  - For category "{category.name}": {comma_and(make_profile(model.accepted_values, boundary_index))}'
    else:
        yield "This is a generic NCS model; sufficient coalitions are specified for each boundary."
        for boundary_index, (category, sufficient_coalitions) in enumerate(zip(problem.ordered_categories[1:], model.sufficient_coalitions)):
            if sufficient_coalitions.is_weights:
                yield f'To get into category "{category.name}", an alternative must be better than the following profile on a set of criteria whose weights add up to at least 1:'
                for profile, weight in zip(make_profile(model.accepted_values, boundary_index), sufficient_coalitions.weights.criterion_weights):
                    yield f'  - {profile} (weight: {weight:.2f})'
            else:
                assert sufficient_coalitions.is_roots
                yield f'The sufficient coalitions for category "{category.name}" are the following, as well as any of their unions:'
                yield from make_upset_roots(sufficient_coalitions.roots.upset_roots)
                yield f'To get into category "{category.name}", an alternative must be better than the following profile on a sufficient coalition of criteria: {comma_and(make_profile(model.accepted_values, boundary_index))}'


class DescribeClassificationModelTestCase(unittest.TestCase):
    maxDiff = None

    problem = Problem(
        [
            Criterion("Criterion 1", Criterion.RealValues(Criterion.PreferenceDirection.increasing, 0, 1)),
            Criterion("Criterion 2", Criterion.RealValues(Criterion.PreferenceDirection.decreasing, 0, 1)),
            Criterion("Criterion 3", Criterion.RealValues(Criterion.PreferenceDirection.increasing, 0, 1)),
            Criterion("Criterion 4", Criterion.RealValues(Criterion.PreferenceDirection.decreasing, 0, 1)),
        ],
        [
            Category("Bad"),
            Category("Intermediate"),
            Category("Good"),
        ],
    )

    def _test(self, model, expected):
        self.assertEqual(list(describe_classification_model(self.problem, model)), expected)

    def test_mrsort(self):
        self._test(
            Model(
                self.problem,
                [
                    AcceptedValues(AcceptedValues.RealThresholds([0.2, 0.7])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.8, 0.7])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.4, 0.5])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.7, 0.3])),
                ],
                [
                    SufficientCoalitions(SufficientCoalitions.Weights([0.7, 0.5, 0.4, 0.2])),
                    SufficientCoalitions(SufficientCoalitions.Weights([0.7, 0.5, 0.4, 0.2])),
                ],
            ),
            [
                'This is a MR-Sort (a.k.a. 1-Uc-NCS) model: an NCS model where the sufficient coalitions are specified using the same criterion weights for all boundaries.',
                'The weights associated to each criterion are:',
                '  - Criterion "Criterion 1": 0.70',
                '  - Criterion "Criterion 2": 0.50',
                '  - Criterion "Criterion 3": 0.40',
                '  - Criterion "Criterion 4": 0.20',
                'To get into an upper category, an alternative must be better than the following profiles on a set of criteria whose weights add up to at least 1:',
                '  - For category "Intermediate": at least 0.20 on criterion "Criterion 1", at most 0.80 on criterion "Criterion 2", at least 0.40 on criterion "Criterion 3", and at most 0.70 on criterion "Criterion 4"',
                '  - For category "Good": at least 0.70 on criterion "Criterion 1", at most 0.70 on criterion "Criterion 2", at least 0.50 on criterion "Criterion 3", and at most 0.30 on criterion "Criterion 4"'
            ],
        )

    def test_ucncs(self):
        self._test(
            Model(
                self.problem,
                [
                    AcceptedValues(AcceptedValues.RealThresholds([0.2, 0.7])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.8, 0.7])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.4, 0.5])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.7, 0.3])),
                ],
                [
                    SufficientCoalitions(SufficientCoalitions.Roots(self.problem, [[0, 1], [0, 2], [1, 2, 3]])),
                    SufficientCoalitions(SufficientCoalitions.Roots(self.problem, [[0, 1], [0, 2], [1, 2, 3]])),
                ],
            ),
            [
                'This is a Uc-NCS model: an NCS model with the same sufficient coalitions for all boundaries.',
                'The sufficient coalitions of criteria are the following, as well as any of their unions:',
                '  - "Criterion 1" and "Criterion 2"',
                '  - "Criterion 1" and "Criterion 3"',
                '  - "Criterion 2", "Criterion 3", and "Criterion 4"',
                'To get into an upper category, an alternative must be better than the following profiles on a sufficient coalition of criteria:',
                '  - For category "Intermediate": at least 0.20 on criterion "Criterion 1", at most 0.80 on criterion "Criterion 2", at least 0.40 on criterion "Criterion 3", and at most 0.70 on criterion "Criterion 4"',
                '  - For category "Good": at least 0.70 on criterion "Criterion 1", at most 0.70 on criterion "Criterion 2", at least 0.50 on criterion "Criterion 3", and at most 0.30 on criterion "Criterion 4"'
            ],
        )

    def test_mixed(self):
        self._test(
            Model(
                self.problem,
                [
                    AcceptedValues(AcceptedValues.RealThresholds([0.2, 0.7])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.8, 0.7])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.4, 0.5])),
                    AcceptedValues(AcceptedValues.RealThresholds([0.7, 0.3])),
                ],
                [
                    SufficientCoalitions(SufficientCoalitions.Roots(self.problem, [[0, 1], [0, 2], [1, 2, 3]])),
                    SufficientCoalitions(SufficientCoalitions.Weights([0.7, 0.5, 0.4, 0.2])),
                ],
            ),
            [
                'This is a generic NCS model; sufficient coalitions are specified for each boundary.',
                'The sufficient coalitions for category "Intermediate" are the following, as well as any of their unions:',
                '  - "Criterion 1" and "Criterion 2"',
                '  - "Criterion 1" and "Criterion 3"',
                '  - "Criterion 2", "Criterion 3", and "Criterion 4"',
                'To get into category "Intermediate", an alternative must be better than the following profile on a sufficient coalition of criteria: at least 0.20 on criterion "Criterion 1", at most 0.80 on criterion "Criterion 2", at least 0.40 on criterion "Criterion 3", and at most 0.70 on criterion "Criterion 4"',
                'To get into category "Good", an alternative must be better than the following profile on a set of criteria whose weights add up to at least 1:',
                '  - at least 0.70 on criterion "Criterion 1" (weight: 0.70)',
                '  - at most 0.70 on criterion "Criterion 2" (weight: 0.50)',
                '  - at least 0.50 on criterion "Criterion 3" (weight: 0.40)',
                '  - at most 0.30 on criterion "Criterion 4" (weight: 0.20)'
            ]
        )



# Copyright 2023-2024 Vincent Jacques

from __future__ import annotations

import contextlib
import json
import math
import os
import random
import sys

import click
import matplotlib.pyplot as plt

import lincs


def options_tree(name, kwds, dependents):
    """
    Ad-hoc decorator for a tree dependent of click.options.
    Options down the tree are only valid if the options up the tree are set to specific values.
    Think of them as sub-options of the previous ones.
    """

    def decorator(command):
        def walk(
            option_name_prefix_,
            parameter_name_prefix_,
            conditions_,
            name_,
            kwds_,
            dependents_,
        ):
            if conditions_:
                # \b prevents rewrapping of the paragraph by Click
                # (https://click.palletsprojects.com/en/8.1.x/api/#click.wrap_text)
                conditions__ = f"\n\n\b\nOnly valid if:\n"
                if os.environ.get("LINCS_GENERATING_SPHINX_DOC") == "1":
                    conditions__ += "\n"
                    quote = "``"
                else:
                    quote = "'"
                for (k, v) in conditions_:
                    conditions__ += f"- {quote}{k}{quote} is {quote}{v}{quote}\n"
                # This non-blocking space ensures Click does not put the default
                # value on the same line as the last condition
                conditions__ += f"\n\n "

            for value__, dependents__ in dependents_.items():
                for name___, kwds___, dependents___ in reversed(dependents__):
                    walk(
                        f"{option_name_prefix_}{value__}.",
                        f"{parameter_name_prefix_}{value__.replace('-', '_')}__",
                        conditions_ + [(f"--{option_name_prefix_}{name_}", value__)],
                        name___,
                        kwds___,
                        dependents___,
                    )

            help = kwds_.pop("help", None)
            if conditions_:
                if help is None:
                    help = conditions__
                else:
                    assert help.endswith(".") or help.endswith(")")
                    help = f"{help} {conditions__}"

            click.option(
                f"--{option_name_prefix_}{name_}",
                f"{parameter_name_prefix_}{name_.replace('-', '_')}",
                help=help,
                **kwds_
            )(
                command,
            )

        walk("", "", [], name, kwds, dependents)

        return command

    return decorator


@click.group(
    help="""
        lincs (Learn and Infer Non-Compensatory Sorting) is a set of tools for training and using MCDA models.
    """,
)
@click.version_option(version=lincs.__version__, message="%(version)s")
def main():
    pass


@main.group(
    help="Get information about lincs itself."
)
def info():
    pass

@info.command(
    help="Check whether lincs was compiled with CUDA support. Return code is 0 if CUDA is supported, 1 otherwise."
)
@click.option(
    "--quiet", is_flag=True,
    help="Don't print anything, just return the exit code.",
)
def has_gpu(quiet):
    if lincs.has_gpu:
        if not quiet:
            print("lincs was compiled with CUDA support")
    else:
        if not quiet:
            print("lincs was compiled WITHOUT CUDA support")
        exit(1)


@main.group(
    help="Generate synthetic data.",
)
def generate():
    pass


@generate.command(
    help="""
        Generate a synthetic classification problem.

        The generated problem has CRITERIA_COUNT criteria and CATEGORIES_COUNT categories.
    """,
)
@click.argument(
    "criteria-count",
    type=click.IntRange(min=1),
)
@click.argument(
    "categories-count",
    type=click.IntRange(min=1),
)
@click.option(
    "--denormalized-min-max",
    is_flag=True,
    help="Generate criteria with random denormalized min and max values. (By default, min and max value are 0 and 1)",
)
@click.option(
    "--forbid-increasing-criteria",
    is_flag=True,
    help="Forbid criteria to have increasing preference direction. (Requires '--allow-decreasing-criteria')",
)
@click.option(
    "--allow-decreasing-criteria",
    is_flag=True,
    help="Allow criteria to have decreasing preference direction. (By default, all criteria have increasing preference direction)",
)
@click.option(
    "--forbid-real-criteria",
    is_flag=True,
    help="Forbid criteria with real values. (Requires another '--allow-...-criteria' option)",
)
@click.option(
    "--allow-enumerated-criteria",
    is_flag=True,
    help="Allow criteria with enumerated values. (By default, all criteria are real)",
)
@click.option(
    "--allow-integer-criteria",
    is_flag=True,
    help="Allow criteria with integer values. (By default, all criteria are real)",
)
@click.option(
    "--output-problem",
    type=click.File(mode="w"),
    default="-",
    help="Write generated problem to this file instead of standard output.",
)
@click.option(
    "--random-seed",
    help="The random seed to use.",
    type=click.IntRange(min=0),
    default=random.randrange(2**30),
)
def classification_problem(
    criteria_count,
    categories_count,
    denormalized_min_max,
    forbid_increasing_criteria, allow_decreasing_criteria,
    forbid_real_criteria, allow_enumerated_criteria, allow_integer_criteria,
    output_problem,
    random_seed
):
    command_line = ["lincs", "generate", "classification-problem", criteria_count, categories_count, "--random-seed", random_seed]
    if denormalized_min_max:
        command_line += ["--denormalized-min-max"]

    allowed_preference_directions = []
    if forbid_increasing_criteria:
        command_line += ["--forbid-increasing-criteria"]
    else:
        allowed_preference_directions.append(lincs.classification.Criterion.PreferenceDirection.increasing)
    if allow_decreasing_criteria:
        command_line += ["--allow-decreasing-criteria"]
        allowed_preference_directions.append(lincs.classification.Criterion.PreferenceDirection.decreasing)

    allowed_value_types = []
    if forbid_real_criteria:
        command_line += ["--forbid-real-criteria"]
    else:
        allowed_value_types.append(lincs.classification.Criterion.ValueType.real)
    if allow_enumerated_criteria:
        command_line += ["--allow-enumerated-criteria"]
        allowed_value_types.append(lincs.classification.Criterion.ValueType.enumerated)
    if allow_integer_criteria:
        command_line += ["--allow-integer-criteria"]
        allowed_value_types.append(lincs.classification.Criterion.ValueType.integer)

    if not allowed_value_types:
        print("ERROR: no allowed value type. '--forbid-real-criteria' requires at least one of '--allow-enumerated-criteria' or '--allow-integer-criteria'", file=sys.stderr)
        print(make_reproduction_command(command_line), file=sys.stderr)
        exit(1)

    if not allowed_preference_directions:
        print("ERROR: no allowed preference direction. '--forbid-increasing-criteria' requires '--allow-decreasing-criteria'", file=sys.stderr)
        print(make_reproduction_command(command_line), file=sys.stderr)
        exit(1)

    print(f"# {make_reproduction_command(command_line)}", file=output_problem, flush=True)

    problem = lincs.classification.generate_problem(
        criteria_count,
        categories_count,
        random_seed=random_seed,
        normalized_min_max=not denormalized_min_max,
        allowed_preference_directions=allowed_preference_directions,
        allowed_value_types=allowed_value_types,
    )
    problem.dump(output_problem)


@generate.command(
    help="""
        Generate a synthetic classification model.

        PROBLEM is a *classification problem* file describing the problem to generate a model for.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.option(
    "--output-model",
    type=click.File(mode="w"),
    default="-",
    help="Write generated model to this file instead of standard output.",
)
@click.option(
    "--random-seed",
    help="The random seed to use.",
    type=click.IntRange(min=0),
    default=random.randrange(2**30),
)
@options_tree(
    "model-type",
    dict(
        help="The type of classification model to generate.",
        type=click.Choice(["mrsort"]),
        default="mrsort",
        show_default=True,
    ),
    {
        "mrsort": [
            (
                "fixed-weights-sum",
                dict(
                    help="Make sure weights add up to this pre-determined value instead of a pseudo-random one.",
                    type=click.FloatRange(min=1.0),
                    default=None,
                    show_default=True,
                ),
                {},
            ),
        ],
    },
)
def classification_model(
    problem,
    output_model,
    random_seed,
    model_type,
    mrsort__fixed_weights_sum,
):
    command_line = ["lincs", "generate", "classification-model", get_input_file_name(problem), "--random-seed", random_seed, "--model-type", model_type]
    if mrsort__fixed_weights_sum is not None:
        command_line += ["--mrsort.fixed-weights-sum", mrsort__fixed_weights_sum]
    print(f"# {make_reproduction_command(command_line)}", file=output_model, flush=True)

    with loading_guard():
        problem = lincs.classification.Problem.load(problem)

    assert model_type == "mrsort"
    model = lincs.classification.generate_mrsort_model(
        problem,
        random_seed=random_seed,
        fixed_weights_sum=mrsort__fixed_weights_sum,
    )
    model.dump(problem, output_model)


@generate.command(
    help="""
        Generate synthetic classified alternatives.

        PROBLEM is a *classification problem* file describing the problem to generate alternatives for.
        MODEL is a *classification model* file for that problem describing the model to use to classify the generated alternatives.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.argument(
    "model",
    type=click.File(mode="r"),
)
@click.argument(
    "alternatives-count",
    type=click.IntRange(min=1),
)
@click.option(
    "--output-alternatives",
    type=click.File(mode="w"),
    default="-",
    help="Write generated classified alternatives to this file instead of standard output.",
)
@click.option(
    "--max-imbalance",
    type=click.FloatRange(min=0.0, max=1.0, max_open=True),
    default=None,
    help="Ensure that categories are balanced, by forcing their size to differ from the perfectly balanced size by at most this fraction.",
)
# @todo(Feature, later) Consider creating a 'lincs misclassify-alternatives' command
# that mimics the 'lincs generate classified-alternatives --misclassified-count' option below,
# but on a pre-existing learning set
@click.option(
    "--misclassified-count",
    type=click.IntRange(min=0),
    default=0,
    help="Misclassify that many alternatives.",
)
@click.option(
    "--random-seed",
    help="The random seed to use.",
    type=click.IntRange(min=0),
    default=random.randrange(2**30),
)
def classified_alternatives(
    problem,
    model,
    alternatives_count,
    output_alternatives,
    max_imbalance,
    misclassified_count,
    random_seed,
):
    command_line = ["lincs", "generate", "classified-alternatives", get_input_file_name(problem), get_input_file_name(model), alternatives_count, "--random-seed", random_seed]
    if max_imbalance is not None:
        command_line += ["--max-imbalance", max_imbalance]
    command_line += ["--misclassified-count", misclassified_count]

    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
        model = lincs.classification.Model.load(problem, model)

    try:
        alternatives = lincs.classification.generate_alternatives(
            problem,
            model,
            alternatives_count,
            random_seed=random_seed,
            max_imbalance=max_imbalance,
        )
    except lincs.classification.BalancedAlternativesGenerationException:
        print("ERROR: lincs is unable to generate a balanced set of classified alternatives. Try to increase the allowed imbalance, or use a more lenient model?", file=sys.stderr)
        print(make_reproduction_command(command_line), file=sys.stderr)
        exit(1)
    else:
        if misclassified_count:
            lincs.classification.misclassify_alternatives(
                problem,
                alternatives,
                misclassified_count,
                random_seed=random_seed + 27,  # Arbitrary, does not hurt
            )
        print(f"# {make_reproduction_command(command_line)}", file=output_alternatives, flush=True)
        alternatives.dump(problem, output_alternatives)


@main.group(
    help="Make graphs from data.",
)
def visualize():
    pass


@visualize.command(
    help="""
        Visualize a classification model.

        PROBLEM is a *classification problem* file.
        MODEL is a *classification model* file for that problem describing the model to visualize.
        The generated image is written to the OUTPUT file in PNG format.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.argument(
    "model",
    type=click.File(mode="r"),
)
@click.option(
    "--alternatives",
    type=click.File(mode="r"),
    help="Add the alternatives from this *classified alternatives* file to the visualization.",
)
@click.option(
    "--alternatives-count",
    type=int,
    help="Add only this number of alternatives.",
)
@click.argument(
    "output",
    type=click.File(mode="wb"),
)
def classification_model(
    problem,
    model,
    alternatives,
    alternatives_count,
    output,
):
    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
        model = lincs.classification.Model.load(problem, model)
        if alternatives is None:
            alternatives = []
        else:
            alternatives = lincs.classification.Alternatives.load(problem, alternatives).alternatives
            if alternatives_count is not None:
                alternatives = alternatives[:alternatives_count]

    figure, axes = plt.subplots(1, 1, figsize=(6, 4), layout="constrained")
    lincs.classification.visualize_model(problem, model, alternatives, axes)
    figure.savefig(output, format="png", dpi=100)
    plt.close(figure)


@main.group(
    help="Provide human-readable descriptions.",
)
def describe():
    pass


@describe.command(
    help="""
        Describe a classification problem.

        PROBLEM is a *classification problem* file.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.option(
    "--output-description",
    type=click.File(mode="w"),
    default="-",
    help="Write description to this file instead of standard output.",
)
def classification_problem(
    problem,
    output_description,
):
    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
    for line in lincs.classification.describe_problem(problem):
        print(line, file=output_description)


@describe.command(
    help="""
        Describe a classification model.

        PROBLEM is a *classification problem* file.
        MODEL is a *classification model* file for that problem.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.argument(
    "model",
    type=click.File(mode="r"),
)
@click.option(
    "--output-description",
    type=click.File(mode="w"),
    default="-",
    help="Write description to this file instead of standard output.",
)
def classification_model(
    problem,
    model,
    output_description,
):
    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
        model = lincs.classification.Model.load(problem, model)
    for line in lincs.classification.describe_model(problem, model):
        print(line, file=output_description)


@main.group(
    help="Learn a model.",
)
def learn():
    pass


max_sat_options = [
    (
        "solver",
        dict(
            help="The solver to use to solve the MaxSAT problem.",
            type=click.Choice(["eval-max-sat"]),
            default="eval-max-sat",
            show_default=True,
        ),
        {
            "eval-max-sat": [
                # These three options correspond to EvalMaxSAT's command-line options here:
                # https://github.com/normal-account/EvalMaxSAT2022/blob/main/main.cpp#L43-L50
                (
                    "nb-minimize-threads",
                    dict(
                        help="The number of threads to use to minimize the MaxSAT problem. Passed directly to the EvalMaxSAT solver.",
                        type=click.IntRange(min=0),
                        default=0,
                        show_default=True,
                    ),
                    {},
                ),
                (
                    "timeout-fast-minimize",
                    dict(
                        help="The maximum duration of the \"fast minimize\" phase of solving the MaxSAT problem, in seconds. Passed directly to the EvalMaxSAT solver.",
                        type=click.IntRange(min=0),
                        default=60,
                        show_default=True,
                    ),
                    {},
                ),
                (
                    "coef-minimize-time",
                    dict(
                        help="The coefficient to use to multiply the time spent minimizing the MaxSAT problem. Passed directly to the EvalMaxSAT solver.",
                        type=click.IntRange(min=0),
                        default=2,
                        show_default=True,
                    ),
                    {},
                ),
            ],
        },
    ),
]

@learn.command(
    help="""
        Learn a classification model.

        PROBLEM is a *classification problem* file describing the problem to learn a model for.
        LEARNING_SET is a *classified alternatives* file for that problem.
        It's used as a source of truth to learn the model.

        If you use the --mrsort.weights-profiles-breed strategy, you SHOULD specify at least one
        termination strategy, e.g. --mrsort.weights-profiles-breed.max-duration.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.argument(
    "learning-set",
    type=click.File(mode="r"),
)
@click.option(
    "--output-model",
    type=click.File(mode="w"),
    default="-",
    help="Write the learned classification model to this file instead of standard output.",
)
@options_tree(
    "model-type",
    dict(
        help="The type of classification model to learn.",
        type=click.Choice(["mrsort", "ucncs"]),
        default="mrsort",
        show_default=True,
    ),
    {
        "mrsort": [
            (
                "strategy",
                dict(
                    help="The top-level strategy to use to learn the MRSort model. See https://mics-lab.github.io/lincs/user-guide.html#learning-strategies about strategies.",
                    type=click.Choice(["weights-profiles-breed"]),
                    default="weights-profiles-breed",
                    show_default=True,
                ),
                {
                    "weights-profiles-breed": [
                        (
                            "target-accuracy",
                            dict(
                                help="The target accuracy to reach on the learning set.",
                                type=click.FloatRange(min=0.0, max=1.0),
                                default=1.0,
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "max-iterations",
                            dict(
                                help="The maximum number of iterations to use to learn the MRSort model.",
                                type=click.IntRange(min=1),
                                default=None,
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "max-iterations-without-progress",
                            dict(
                                help="The maximum number of iterations to try learning the MRSort model without progressing before giving up.",
                                type=click.IntRange(min=1),
                                default=None,
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "max-duration",
                            dict(
                                help="The maximum duration to learn the MRSort model, in seconds.",
                                type=click.FloatRange(min=0),
                                default=None,
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "max-duration-without-progress",
                            dict(
                                help="The maximum duration to try learning the MRSort model without progressing before giving up, in seconds.",
                                type=click.FloatRange(min=0),
                                default=None,
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "models-count",
                            dict(
                                help="The number of temporary MRSort models to train. The result of the learning will be the most accurate of those models.",
                                type=click.IntRange(min=1),
                                default=9,
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "initialization-strategy",
                            dict(
                                help="The strategy to use to initialize the MRSort models.",
                                type=click.Choice(["maximize-discrimination-per-criterion"]),
                                default="maximize-discrimination-per-criterion",
                                show_default=True,
                            ),
                            {},
                        ),
                        (
                            "weights-strategy",
                            dict(
                                help="The strategy to use to improve the weights of the MRSort models.",
                                type=click.Choice(["linear-program"]),
                                default="linear-program",
                                show_default=True,
                            ),
                            {
                                "linear-program": [
                                    (
                                        "solver",
                                        dict(
                                            help="The solver to use to solve the linear programs.",
                                            type=click.Choice(["glop", "alglib"]),
                                            default="glop",
                                            show_default=True,
                                        ),
                                        {},
                                    ),
                                ],
                            },
                        ),
                        (
                            "profiles-strategy",
                            dict(
                                help="The strategy to use to improve the profiles of the MRSort models.",
                                type=click.Choice(["accuracy-heuristic"]),
                                default="accuracy-heuristic",
                                show_default=True,
                            ),
                            {
                                "accuracy-heuristic": [
                                    (
                                        "random-seed",
                                        dict(
                                            help="The random seed to use for this heuristic.",
                                            type=click.IntRange(min=0),
                                            default=random.randrange(2**30),
                                        ),
                                        {},
                                    ),
                                    (
                                        "processor",
                                        dict(
                                            help="The processor to use to improve the profiles of the MRSort models."
                                            + ("" if lincs.has_gpu else " (Only 'cpu' is available because lincs was compiled without 'nvcc')"),
                                            type=click.Choice(["cpu"] + (["gpu"] if lincs.has_gpu else [])),
                                            default="cpu",
                                            show_default=True,
                                        ),
                                        {},
                                    ),
                                ],
                            },
                        ),
                        (
                            "breed-strategy",
                            dict(
                                help="The strategy to use to breed the MRSort models.",
                                type=click.Choice(["reinitialize-least-accurate"]),
                                default="reinitialize-least-accurate",
                                show_default=True,
                            ),
                            {
                                "reinitialize-least-accurate": [
                                    (
                                        "portion",
                                        dict(
                                            help="The portion of the least accurate MRSort models to reinitialize.",
                                            type=click.FloatRange(min=0.0, max=1.0),
                                            default=0.5,
                                            show_default=True,
                                        ),
                                        {},
                                    ),
                                ],
                            },
                        ),
                        (
                            "verbose",
                            dict(
                                help="Print information about the learning process on stderr while learning.",
                                is_flag=True,
                            ),
                            {},
                        ),
                        (
                            "output-metadata",
                            dict(
                                help="Write metadata about the learning process to this file.",
                                type=click.File(mode="w"),
                            ),
                            {},
                        )
                    ],
                },
            ),
        ],
        "ucncs": [
            (
                "strategy",
                dict(
                    help="The general approach to transform the learning problem into a satisfiability problem.",
                    type=click.Choice(["sat-by-coalitions", "sat-by-separation", "max-sat-by-coalitions", "max-sat-by-separation"]),
                    default="sat-by-coalitions",
                    show_default=True,
                ),
                {
                    "max-sat-by-coalitions": max_sat_options,
                    "max-sat-by-separation": max_sat_options,
                },
            ),
        ],
    },
)
def classification_model(
    problem,
    learning_set,
    output_model,
    model_type,
    mrsort__strategy,
    mrsort__weights_profiles_breed__target_accuracy,
    mrsort__weights_profiles_breed__max_iterations,
    mrsort__weights_profiles_breed__max_iterations_without_progress,
    mrsort__weights_profiles_breed__max_duration,
    mrsort__weights_profiles_breed__max_duration_without_progress,
    mrsort__weights_profiles_breed__models_count,
    mrsort__weights_profiles_breed__initialization_strategy,
    mrsort__weights_profiles_breed__weights_strategy,
    mrsort__weights_profiles_breed__linear_program__solver,
    mrsort__weights_profiles_breed__profiles_strategy,
    mrsort__weights_profiles_breed__accuracy_heuristic__random_seed,
    mrsort__weights_profiles_breed__accuracy_heuristic__processor,
    mrsort__weights_profiles_breed__breed_strategy,
    mrsort__weights_profiles_breed__reinitialize_least_accurate__portion,
    mrsort__weights_profiles_breed__verbose,
    mrsort__weights_profiles_breed__output_metadata,
    ucncs__strategy,
    ucncs__max_sat_by_coalitions__solver,
    ucncs__max_sat_by_coalitions__eval_max_sat__nb_minimize_threads,
    ucncs__max_sat_by_coalitions__eval_max_sat__timeout_fast_minimize,
    ucncs__max_sat_by_coalitions__eval_max_sat__coef_minimize_time,
    ucncs__max_sat_by_separation__solver,
    ucncs__max_sat_by_separation__eval_max_sat__nb_minimize_threads,
    ucncs__max_sat_by_separation__eval_max_sat__timeout_fast_minimize,
    ucncs__max_sat_by_separation__eval_max_sat__coef_minimize_time,
):
    command_line = ["lincs", "learn", "classification-model", get_input_file_name(problem), get_input_file_name(learning_set), "--model-type", model_type]

    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
        learning_set = lincs.classification.Alternatives.load(problem, learning_set)

    if model_type == "mrsort":
        command_line += ["--mrsort.strategy", mrsort__strategy]
        if mrsort__strategy == "weights-profiles-breed":
            command_line += ["--mrsort.weights-profiles-breed.models-count", mrsort__weights_profiles_breed__models_count]
            command_line += ["--mrsort.weights-profiles-breed.accuracy-heuristic.random-seed", mrsort__weights_profiles_breed__accuracy_heuristic__random_seed]

            learning_data = lincs.classification.LearnMrsortByWeightsProfilesBreed.LearningData(problem, learning_set, mrsort__weights_profiles_breed__models_count, mrsort__weights_profiles_breed__accuracy_heuristic__random_seed)

            command_line += ["--mrsort.weights-profiles-breed.initialization-strategy", mrsort__weights_profiles_breed__initialization_strategy]
            if mrsort__weights_profiles_breed__initialization_strategy == "maximize-discrimination-per-criterion":
                profiles_initialization_strategy = lincs.classification.InitializeProfilesForProbabilisticMaximalDiscriminationPowerPerCriterion(learning_data)

            command_line += ["--mrsort.weights-profiles-breed.weights-strategy", mrsort__weights_profiles_breed__weights_strategy]
            if mrsort__weights_profiles_breed__weights_strategy == "linear-program":
                command_line += ["--mrsort.weights-profiles-breed.linear-program.solver", mrsort__weights_profiles_breed__linear_program__solver]
                if mrsort__weights_profiles_breed__linear_program__solver == "glop":
                    weights_optimization_strategy = lincs.classification.OptimizeWeightsUsingGlop(learning_data)
                elif mrsort__weights_profiles_breed__linear_program__solver == "alglib":
                    weights_optimization_strategy = lincs.classification.OptimizeWeightsUsingAlglib(learning_data)

            command_line += ["--mrsort.weights-profiles-breed.profiles-strategy", mrsort__weights_profiles_breed__profiles_strategy]
            if mrsort__weights_profiles_breed__profiles_strategy == "accuracy-heuristic":
                command_line += ["--mrsort.weights-profiles-breed.accuracy-heuristic.processor", mrsort__weights_profiles_breed__accuracy_heuristic__processor]
                if mrsort__weights_profiles_breed__accuracy_heuristic__processor == "cpu":
                    profiles_improvement_strategy = lincs.classification.ImproveProfilesWithAccuracyHeuristicOnCpu(learning_data)
                elif mrsort__weights_profiles_breed__accuracy_heuristic__processor == "gpu":
                    assert lincs.has_gpu
                    profiles_improvement_strategy = lincs.classification.ImproveProfilesWithAccuracyHeuristicOnGpu(learning_data)

            command_line += ["--mrsort.weights-profiles-breed.breed-strategy", mrsort__weights_profiles_breed__breed_strategy]
            if mrsort__weights_profiles_breed__breed_strategy == "reinitialize-least-accurate":
                command_line += ["--mrsort.weights-profiles-breed.reinitialize-least-accurate.portion", mrsort__weights_profiles_breed__reinitialize_least_accurate__portion]
                count = int(mrsort__weights_profiles_breed__reinitialize_least_accurate__portion * mrsort__weights_profiles_breed__models_count)
                breeding_strategy = lincs.classification.ReinitializeLeastAccurate(learning_data, profiles_initialization_strategy, count)

            command_line += ["--mrsort.weights-profiles-breed.target-accuracy", mrsort__weights_profiles_breed__target_accuracy]
            termination_strategies = [lincs.classification.TerminateAtAccuracy(
                learning_data,
                math.ceil(mrsort__weights_profiles_breed__target_accuracy * len(learning_set.alternatives)),
            )]
            if mrsort__weights_profiles_breed__max_iterations is not None:
                command_line += ["--mrsort.weights-profiles-breed.max-iterations", mrsort__weights_profiles_breed__max_iterations]
                termination_strategies.append(lincs.classification.TerminateAfterIterations(learning_data, mrsort__weights_profiles_breed__max_iterations))
            if mrsort__weights_profiles_breed__max_iterations_without_progress is not None:
                command_line += ["--mrsort.weights-profiles-breed.max-iterations-without-progress", mrsort__weights_profiles_breed__max_iterations_without_progress]
                termination_strategies.append(lincs.classification.TerminateAfterIterationsWithoutProgress(learning_data, mrsort__weights_profiles_breed__max_iterations_without_progress))
            if mrsort__weights_profiles_breed__max_duration is not None:
                command_line += ["--mrsort.weights-profiles-breed.max-duration", mrsort__weights_profiles_breed__max_duration]
                termination_strategies.append(lincs.classification.TerminateAfterSeconds(mrsort__weights_profiles_breed__max_duration))
            if mrsort__weights_profiles_breed__max_duration_without_progress is not None:
                command_line += ["--mrsort.weights-profiles-breed.max-duration-without-progress", mrsort__weights_profiles_breed__max_duration_without_progress]
                termination_strategies.append(lincs.classification.TerminateAfterSecondsWithoutProgress(mrsort__weights_profiles_breed__max_duration_without_progress))
            if len(termination_strategies) == 1:
                termination_strategy = termination_strategies[0]
            else:
                termination_strategy = lincs.classification.TerminateWhenAny(termination_strategies)

            observers = []
            if mrsort__weights_profiles_breed__verbose:
                class VerboseObserver(lincs.classification.LearnMrsortByWeightsProfilesBreed.Observer):
                    def __init__(self, learning_data):
                        super().__init__()
                        self.learning_data = learning_data

                    def after_iteration(self):
                        print(f"Best accuracy (after {self.learning_data.iteration_index + 1} iterations): {self.learning_data.get_best_accuracy()}", file=sys.stderr)

                    def before_return(self):
                        print(f"Final accuracy (after {self.learning_data.iteration_index + 1} iterations): {self.learning_data.get_best_accuracy()}", file=sys.stderr)

                observers.append(VerboseObserver(learning_data))
            if mrsort__weights_profiles_breed__output_metadata:
                class MetadataObserver(lincs.classification.LearnMrsortByWeightsProfilesBreed.Observer):
                    def __init__(self, learning_data):
                        super().__init__()
                        self.learning_data = learning_data
                        self.accuracies = []

                    def after_iteration(self):
                        self.accuracies.append(self.learning_data.get_best_accuracy())

                    def before_return(self):
                        self.accuracies.append(self.learning_data.get_best_accuracy())

                metadata_observer = MetadataObserver(learning_data)
                observers.append(metadata_observer)

            learning = lincs.classification.LearnMrsortByWeightsProfilesBreed(
                learning_data,
                profiles_initialization_strategy,
                weights_optimization_strategy,
                profiles_improvement_strategy,
                breeding_strategy,
                termination_strategy,
                observers,
            )
    elif model_type == "ucncs":
        command_line += ["--ucncs.strategy", ucncs__strategy]
        if ucncs__strategy == "sat-by-coalitions":
            learning = lincs.classification.LearnUcncsBySatByCoalitionsUsingMinisat(problem, learning_set)
        elif ucncs__strategy == "sat-by-separation":
            learning = lincs.classification.LearnUcncsBySatBySeparationUsingMinisat(problem, learning_set)
        elif ucncs__strategy == "max-sat-by-coalitions":
            command_line += ["--ucncs.max-sat-by-coalitions.solver", ucncs__max_sat_by_coalitions__solver]
            if ucncs__max_sat_by_coalitions__solver == "eval-max-sat":
                command_line += [
                    "--ucncs.max-sat-by-coalitions.eval-max-sat.nb-minimize-threads", ucncs__max_sat_by_coalitions__eval_max_sat__nb_minimize_threads,
                    "--ucncs.max-sat-by-coalitions.eval-max-sat.timeout-fast-minimize", ucncs__max_sat_by_coalitions__eval_max_sat__timeout_fast_minimize,
                    "--ucncs.max-sat-by-coalitions.eval-max-sat.coef-minimize-time", ucncs__max_sat_by_coalitions__eval_max_sat__coef_minimize_time,
                ]
                learning = lincs.classification.LearnUcncsByMaxSatByCoalitionsUsingEvalmaxsat(
                    problem,
                    learning_set,
                    ucncs__max_sat_by_coalitions__eval_max_sat__nb_minimize_threads,
                    ucncs__max_sat_by_coalitions__eval_max_sat__timeout_fast_minimize,
                    ucncs__max_sat_by_coalitions__eval_max_sat__coef_minimize_time,
                )
        elif ucncs__strategy == "max-sat-by-separation":
            command_line += ["--ucncs.max-sat-by-separation.solver", ucncs__max_sat_by_separation__solver]
            if ucncs__max_sat_by_separation__solver == "eval-max-sat":
                command_line += [
                    "--ucncs.max-sat-by-separation.eval-max-sat.nb-minimize-threads", ucncs__max_sat_by_separation__eval_max_sat__nb_minimize_threads,
                    "--ucncs.max-sat-by-separation.eval-max-sat.timeout-fast-minimize", ucncs__max_sat_by_separation__eval_max_sat__timeout_fast_minimize,
                    "--ucncs.max-sat-by-separation.eval-max-sat.coef-minimize-time", ucncs__max_sat_by_separation__eval_max_sat__coef_minimize_time,
                ]
                learning = lincs.classification.LearnUcncsByMaxSatBySeparationUsingEvalmaxsat(
                    problem,
                    learning_set,
                    ucncs__max_sat_by_separation__eval_max_sat__nb_minimize_threads,
                    ucncs__max_sat_by_separation__eval_max_sat__timeout_fast_minimize,
                    ucncs__max_sat_by_separation__eval_max_sat__coef_minimize_time,
                )

    try:
        model = learning.perform()
    except lincs.LearningFailureException:
        print("ERROR: lincs is unable to learn from this learning set using this algorithm and these parameters.", file=sys.stderr)
        print(make_reproduction_command(command_line), file=sys.stderr)
        exit(1)
    else:
        print(f"# {make_reproduction_command(command_line)}", file=output_model, flush=True)
        if model_type == "mrsort" and mrsort__strategy == "weights-profiles-breed" and mrsort__weights_profiles_breed__output_metadata is not None:
            for termination_strategy in termination_strategies:
                if termination_strategy.terminate():
                    termination_condition = {
                        lincs.classification.TerminateAtAccuracy: "target accuracy reached",
                        lincs.classification.TerminateAfterIterations: "maximum total number of iterations reached",
                        lincs.classification.TerminateAfterIterationsWithoutProgress: "maximum number of iterations without progress reached",
                        lincs.classification.TerminateAfterSeconds: "maximum total duration reached",
                        lincs.classification.TerminateAfterSecondsWithoutProgress: "maximum duration without progress reached",
                    }.get(
                        termination_strategy.__class__,
                        f"{termination_strategy.__class__.__name__} (Unexpected, please let the lincs maintainers know about this)"
                    )
                    break
            else:
                termination_condition = "unknown (Unexpected, please let the lincs maintainers know about this)"
            json.dump(
                {
                    "termination_condition": termination_condition,
                    "iterations_count": learning_data.iteration_index,
                    "intermediate_accuracies": metadata_observer.accuracies,
                },
                mrsort__weights_profiles_breed__output_metadata,
                indent=4
            )
        model.dump(problem, output_model)


@main.command(
    help="""
        Classify alternatives.

        PROBLEM is a *classification problem* file.
        MODEL is a *classification model* file for that problem.
        ALTERNATIVES is an *unclassified alternatives* file for that problem.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.argument(
    "model",
    type=click.File(mode="r"),
)
@click.argument(
    "alternatives",
    type=click.File(mode="r"),
)
@click.option(
    "--output-alternatives",
    type=click.File(mode="w"),
    default="-",
    help="Write classified alternatives to this file instead of standard output.",
)
def classify(
    problem,
    model,
    alternatives,
    output_alternatives,
):
    command_line = ["lincs", "classify", get_input_file_name(problem), get_input_file_name(model), get_input_file_name(alternatives)]

    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
        model = lincs.classification.Model.load(problem, model)
        alternatives = lincs.classification.Alternatives.load(problem, alternatives)

    lincs.classification.classify_alternatives(problem, model, alternatives)
    print(f"# {make_reproduction_command(command_line)}", file=output_alternatives, flush=True)
    alternatives.dump(problem, output_alternatives)


@main.command(
    help="""
        Compute a classification accuracy.

        PROBLEM is a *classification problem* file.
        MODEL is a *classification model* file for that problem.
        TESTING_SET is a *classified alternatives* file for that problem.

        The classification accuracy is written to standard output as an integer between 0 and the number of alternatives.
    """,
)
@click.argument(
    "problem",
    type=click.File(mode="r"),
)
@click.argument(
    "model",
    type=click.File(mode="r"),
)
@click.argument(
    "testing-set",
    type=click.File(mode="r"),
)
def classification_accuracy(
    problem,
    model,
    testing_set,
):
    with loading_guard():
        problem = lincs.classification.Problem.load(problem)
        model = lincs.classification.Model.load(problem, model)
        testing_set = lincs.classification.Alternatives.load(problem, testing_set)

    result = lincs.classification.classify_alternatives(problem, model, testing_set)
    print(f"{result.unchanged}/{result.changed + result.unchanged}")


@contextlib.contextmanager
def loading_guard():
    try:
        yield
    except lincs.DataValidationException as e:
        print(f"ERROR: lincs found an issue with the input files: {e}", file=sys.stderr)
        exit(1)


def get_input_file_name(input_file):
    if input_file.name == "<stdin>":
        return "-"
    return input_file.name


def make_reproduction_command(command):
    return f"Reproduction command (with lincs version {lincs.__version__}): {' '.join(str(c) for c in command)}"
